// Procedural brick texture; brick 4
// inspired by ch. 6 in the orange book
// GZ, 2014, 2021
// http://shaderfrog.com/app/view/5297


precision highp float;
precision highp int;

// good values of the following noise parameters depend heavily on the type of noise function!
uniform float noise_spatial_freq;	// spatial frequency of the noise; start with 6
uniform float noise_amplitude;		// 0.2
uniform float noise_shift;		    // doesn't make sense here, but included for demo purposes

// new: inclue random black spots
uniform float spots_spatial_freq;	// frequency of black spots;  start with 80
uniform float spots_size;			// must be in [0,1], higher = smaller; start 0.7

uniform vec2 BrickStepSize;         // MUST NOT BE 0! use starting value = 0.2 , 0.2
uniform vec2 BrickPercent;          // 0.9, 0.8
uniform vec3 BrickColor;            // starting value = (1.0, 0.1, 0.1)
uniform vec3 MortarColor;           // 0.7, 0.7, 0.7
// DO NOT use declarations like:  uniform vec2 BrickStepSize, BrickPercent;  !!

uniform float ambient;              // ambient light;  start: 0.5

// Default uniforms provided by ShaderFrog
uniform float time;

// Default THREE.js uniforms available to both fragment and vertex shader
uniform mat4 modelMatrix;
uniform mat4 modelViewMatrix;
uniform mat4 projectionMatrix;
uniform mat4 viewMatrix;
uniform mat3 normalMatrix;

//  varyings passed from the vertex shader
varying vec3 vPosition_model;     // the position of the vertex/fragment in model(!) coords
varying vec3 vNormal_model;
varying vec3 vPosition_world;
varying vec3 vNormal_world;

varying float brightness;



// Perlin noise function (see below)
float cnoise(vec2 P);



void main()
{
    vec3 color;
        
	// new: black spots
	if ( cnoise( spots_spatial_freq * vPosition_model.xy + vec2(noise_shift) ) > spots_size )
	{
		// we're inside a black spot
		color = vec3( 0.1, 0.1, 0.1 );
	}
	else
	{

    	vec2 rel_pos = vPosition_model.xy / BrickStepSize;	  // now pos in [k,k+1] for one brick

        // are we in an odd brick row?
        if ( mod(abs(rel_pos.y), 2.0) > 1.0 )           // this saves conversion to int / mod
            rel_pos.x += 0.5;

        rel_pos = fract( rel_pos );                     // = rel. position within brick
        
        vec3 dirt;
    	if ( any( greaterThan(rel_pos, BrickPercent) ) )
    	{
    		color = MortarColor;
    		// add some "dirt"
    		float c = noise_amplitude * cnoise( noise_spatial_freq * vPosition_model.xy + 
    									  vec2(noise_shift) );
    		dirt = vec3( c, c, c );
    	}
    	else
    	{
    		color = BrickColor;
    		// add some "dirt"
    		float c = noise_amplitude * cnoise( noise_spatial_freq * vPosition_model.xy + 
    									  vec2(noise_shift) );
    		dirt = vec3( c, c, c );
    	}
    	color += dirt;
	}
	
    color *= clamp( brightness, 0.0, 1.0);
    gl_FragColor = vec4( color , 1.0);
}




// GLSL textureless classic 2D noise "cnoise",
// https://github.com/stegu/webgl-noise

vec4 mod289(vec4 x)
{
  return x - floor(x * (1.0 / 289.0)) * 289.0;
}

vec4 permute(vec4 x)
{
  return mod289(((x*34.0)+1.0)*x);
}

vec4 taylorInvSqrt(vec4 r)
{
  return 1.79284291400159 - 0.85373472095314 * r;
}

vec2 fade(vec2 t) {
  return t*t*t*(t*(t*6.0-15.0)+10.0);
}

// Classic Perlin noise
float cnoise(vec2 P)
{
  vec4 Pi = floor(P.xyxy) + vec4(0.0, 0.0, 1.0, 1.0);
  vec4 Pf = fract(P.xyxy) - vec4(0.0, 0.0, 1.0, 1.0);
  Pi = mod289(Pi); // To avoid truncation effects in permutation
  vec4 ix = Pi.xzxz;
  vec4 iy = Pi.yyww;
  vec4 fx = Pf.xzxz;
  vec4 fy = Pf.yyww;

  vec4 i = permute(permute(ix) + iy);

  vec4 gx = fract(i * (1.0 / 41.0)) * 2.0 - 1.0 ;
  vec4 gy = abs(gx) - 0.5 ;
  vec4 tx = floor(gx + 0.5);
  gx = gx - tx;

  vec2 g00 = vec2(gx.x,gy.x);
  vec2 g10 = vec2(gx.y,gy.y);
  vec2 g01 = vec2(gx.z,gy.z);
  vec2 g11 = vec2(gx.w,gy.w);

  vec4 norm = taylorInvSqrt(vec4(dot(g00, g00), dot(g01, g01), dot(g10, g10), dot(g11, g11)));
  g00 *= norm.x;  
  g01 *= norm.y;  
  g10 *= norm.z;  
  g11 *= norm.w;  

  float n00 = dot(g00, vec2(fx.x, fy.x));
  float n10 = dot(g10, vec2(fx.y, fy.y));
  float n01 = dot(g01, vec2(fx.z, fy.z));
  float n11 = dot(g11, vec2(fx.w, fy.w));

  vec2 fade_xy = fade(Pf.xy);
  vec2 n_x = mix(vec2(n00, n01), vec2(n10, n11), fade_xy.x);
  float n_xy = mix(n_x.x, n_x.y, fade_xy.y);
  return 2.3 * n_xy;
}



